@@ -1,48 +1,55 @@
 ///////////////////////////////////////////////////////////////////////////
-//
-// Copyright (c) 2006 Academy of Motion Picture Arts and Sciences
+// Copyright (c) 2013 Academy of Motion Picture Arts and Sciences 
 // ("A.M.P.A.S."). Portions contributed by others as indicated.
 // All rights reserved.
 // 
-// A world-wide, royalty-free, non-exclusive right to distribute, copy,
-// modify, create derivatives, and use, in source and binary forms, is
-// hereby granted, subject to acceptance of this license. Performance of
-// any of the aforementioned acts indicates acceptance to be bound by the
-// following terms and conditions:
-// 
-//   * Redistributions of source code must retain the above copyright
-//     notice, this list of conditions and the Disclaimer of Warranty.
-// 
-//   * Redistributions in binary form must reproduce the above copyright
-//     notice, this list of conditions and the Disclaimer of Warranty
-//     in the documentation and/or other materials provided with the
-//     distribution.
-// 
-//   * Nothing in this license shall be deemed to grant any rights to
-//     trademarks, copyrights, patents, trade secrets or any other
-//     intellectual property of A.M.P.A.S. or any contributors, except
-//     as expressly stated herein, and neither the name of A.M.P.A.S.
-//     nor of any other contributors to this software, may be used to
-//     endorse or promote products derived from this software without
-//     specific prior written permission of A.M.P.A.S. or contributor,
-//     as appropriate.
-// 
-// This license shall be governed by the laws of the State of California,
-// and subject to the jurisdiction of the courts therein.
-// 
-// Disclaimer of Warranty: THIS SOFTWARE IS PROVIDED BY A.M.P.A.S. AND
-// CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
-// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
-// FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO
-// EVENT SHALL A.M.P.A.S., ANY CONTRIBUTORS OR DISTRIBUTORS BE LIABLE FOR
-// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
-// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
-// IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
-// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
-// IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+// A worldwide, royalty-free, non-exclusive right to copy, modify, create
+// derivatives, and use, in source and binary forms, is hereby granted, 
+// subject to acceptance of this license. Performance of any of the 
+// aforementioned acts indicates acceptance to be bound by the following 
+// terms and conditions:
 //
+//  * Copies of source code, in whole or in part, must retain the 
+//    above copyright notice, this list of conditions and the 
+//    Disclaimer of Warranty.
+//
+//  * Use in binary form must retain the above copyright notice, 
+//    this list of conditions and the Disclaimer of Warranty in the
+//    documentation and/or other materials provided with the distribution.
+//
+//  * Nothing in this license shall be deemed to grant any rights to 
+//    trademarks, copyrights, patents, trade secrets or any other 
+//    intellectual property of A.M.P.A.S. or any contributors, except 
+//    as expressly stated herein.
+//
+//  * Neither the name "A.M.P.A.S." nor the name of any other 
+//    contributors to this software may be used to endorse or promote 
+//    products derivative of or based on this software without express 
+//    prior written permission of A.M.P.A.S. or the contributors, as 
+//    appropriate.
+// 
+// This license shall be construed pursuant to the laws of the State of 
+// California, and any disputes related thereto shall be subject to the 
+// jurisdiction of the courts therein.
+//
+// Disclaimer of Warranty: THIS SOFTWARE IS PROVIDED BY A.M.P.A.S. AND 
+// CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
+// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS 
+// FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO 
+// EVENT SHALL A.M.P.A.S., OR ANY CONTRIBUTORS OR DISTRIBUTORS, BE LIABLE 
+// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, RESITUTIONARY, 
+// OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
+// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
+// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
+// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
+// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF 
+// THE POSSIBILITY OF SUCH DAMAGE.
+//
+// WITHOUT LIMITING THE GENERALITY OF THE FOREGOING, THE ACADEMY 
+// SPECIFICALLY DISCLAIMS ANY REPRESENTATIONS OR WARRANTIES WHATSOEVER 
+// RELATED TO PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS IN THE ACADEMY 
+// COLOR ENCODING SYSTEM, OR APPLICATIONS THEREOF, HELD BY PARTIES OTHER 
+// THAN A.M.P.A.S., WHETHER DISCLOSED OR UNDISCLOSED.
 ///////////////////////////////////////////////////////////////////////////
 
 
@@ -74,6 +81,7 @@
 #include <CtlAddr.h>
 #include <string>
 #include <vector>
+#include <stdarg.h>
 
 namespace Ctl {
 
@@ -122,6 +130,18 @@ typedef RcPtr <StructType> StructTypePtr;
 class FunctionType;
 typedef RcPtr <FunctionType> FunctionTypePtr;
 
+enum CDataType_e {
+	VoidTypeEnum=0,
+	BoolTypeEnum=1,
+	IntTypeEnum=2,
+	UIntTypeEnum=3,
+	HalfTypeEnum=4,
+	FloatTypeEnum=5,
+	StringTypeEnum=6,
+	StructTypeEnum=7,
+	ArrayTypeEnum=8
+};
+
 typedef std::vector<size_t> SizeVector;
 typedef std::vector<size_t>::iterator SizeVectorIterator;
 typedef std::vector<size_t>::reverse_iterator SizeVectorRIterator;
@@ -130,7 +150,6 @@ class LContext;
 class XContext;
 class Module;
 
-
 class Type: public RcObject
 {
   public:
@@ -200,7 +219,7 @@ class Type: public RcObject
     virtual ExprNodePtr		castValue (LContext &lcontext,
 					   const ExprNodePtr &expr) const = 0;
 
-
+	
     //--------------------------------
     // Print this type (for debugging)
     //--------------------------------
@@ -229,13 +248,42 @@ class Type: public RcObject
 
     virtual void		generateCode (const SyntaxNodePtr &node,
 					      LContext &lcontext) const = 0;
+
+	// This is something of a cheat and mostly for internal use
+	// to avoid doing RTTI for determining type conversions.
+	virtual CDataType_e cDataType() const { return VoidTypeEnum; };
+
+	// These methods are for dealing with complex and aggregate
+	// type introspection. They have the concept of a 'path' where
+	// that consists of a '/' separated list of element identifiers.
+	// In the case of an array an element identifier is interpreted
+	// as a number. In the case of a struct an element identifier is
+	// interpreted as a structure item name. The following element
+	// identifiers have special meaning (but do not conflict with CTL
+	// syntax:
+	//
+	//   *  A null identifier or "." is a no-op to the parser. 
+	//   *  If an identifier is "%s" then the next item on the var-args
+	//      list is interpreted as a char * string and used as the element
+	//      identifier.
+	//   *  If an identifier is "%u", "%d" then the next item on the vargs
+	//      list is interpreted as an integer and used as the element
+	//      identifier.
+	//   *  If an identifier is "%v" then the next item on the vargs list
+	//      is interpreted as a SizeVector * and all elements are used to
+	//      identify an element in a matrix.
+	//
+	static void         childElement(size_t *offset, TypePtr *type,
+	                                 const std::string &path, ...);
+
+	static void         childElementV(size_t *offset, TypePtr *type,
+	                                  const std::string &path, va_list ap);
 };
 
 
 class DataType: public Type
 {
   public:
-
     //------------------------------------------------------------
     // Size and required alignment of objects of this data type:
     //
@@ -249,19 +297,19 @@ class DataType: public Type
     // such that y >= x and y is divisible by objectAlignment().
     //------------------------------------------------------------
 
-    virtual size_t		objectSize () const = 0;
-    virtual size_t		alignedObjectSize () const = 0;
-    virtual size_t		objectAlignment () const = 0;
-    char *			alignObjectAddr (char *addr) const;
+    virtual size_t  objectSize() const=0;
+    virtual size_t  alignedObjectSize() const=0;
+    virtual size_t  objectAlignment() const=0;
+    char *          alignObjectAddr(char *addr) const;
 
     //-------------------------------------------------------
     // For basic data types, appends alignedOjectSize.  For arrays 
     // and structs recurses through sub-elements to return number of 
     // non-struct, non-array elements.
     //-------------------------------------------------------
-    virtual void                coreSizes(size_t parentOffset,
-					  SizeVector &sizes, 
-					  SizeVector &offsets) const;
+    virtual void    coreSizes(size_t parentOffset,
+					          SizeVector &sizes, 
+	                          SizeVector &offsets) const;
 
 
     //-------------------------------------------------------
@@ -287,6 +335,8 @@ class VoidType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return VoidTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
@@ -303,7 +353,6 @@ class VoidType: public DataType
 				      LContext &lcontext) const;
 };
 
-
 class BoolType: public DataType
 {
   public:
@@ -315,6 +364,8 @@ class BoolType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return IntTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
@@ -337,6 +388,8 @@ class IntType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return IntTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
@@ -359,6 +412,8 @@ class UIntType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return UIntTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
@@ -381,6 +436,8 @@ class HalfType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return HalfTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
@@ -403,6 +460,8 @@ class FloatType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return FloatTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
  					  const ExprNodePtr &expr) const;
 
@@ -425,6 +484,8 @@ class StringType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return StringTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
  					  const ExprNodePtr &expr) const;
 
@@ -439,7 +500,6 @@ class StringType: public DataType
 class ArrayType: public DataType
 {
   public:
-
     ArrayType (const DataTypePtr &elementType, int size);
 
     //---------------------------------------------------------------
@@ -459,6 +519,8 @@ class ArrayType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return ArrayTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
@@ -485,9 +547,8 @@ class ArrayType: public DataType
 
   private:
 
-    DataTypePtr			_elementType;
-    int 			_size;
-
+    DataTypePtr _elementType;
+    int         _size;
 };
 
 
@@ -525,6 +586,8 @@ class StructType: public DataType
     virtual bool		canPromoteFrom (const TypePtr &t) const;
     virtual bool		canCastFrom (const TypePtr &t) const;
 
+	virtual CDataType_e cDataType() const { return StructTypeEnum; };
+
     virtual ExprNodePtr		evaluate (LContext &lcontext,
 					  const ExprNodePtr &expr) const;
 
